/*
 * LISTDEFUN.E - list-functions - original author unknown.
 * Lists all C functions in a buffer to a window.
 *
 * Additions by Aaron A. Collins to perform the ever-useful "make
 * prototypes" function on a buffer.  Command make-prototypes performs
 * this, and copies the function prototypes that are being built to a
 * buffer named *Prototypes*.  This is then put into a region and killed,
 * then yanked, so a subsequent yank command will paste the region
 * of prototypes that was built into the current buffer.
 *
 * This may not handle all of the really obfuscated "far function returning
 * a pointer to an array of pointers to arrays of char" or the like, but
 * most if not all functions will be parsed properly, regardless of whether
 * or not the ANSI function decl:
 *
 * ret func(type arg1, type arg2, etc.)
 * {...
 *
 * or the classic k&r function decl:
 *
 * ret func(arg1, arg2, arg3)
 * type arg1, arg2;
 * type arg3;
 * {...
 *
 * format is used.
 *
 * The functionality of this program was inspired by a public domain program
 * called MKPROTO - this code is in no way connected with that one, except
 * that it works the same!
 *
 * Revision History
 * ----------------
 * Version 1.0 - original author unknown/undocumented
 * Version 2.0 - AAC - 07/20/1993 - added make-prototypes command
 *
 */

#include "eel.h"

#define MAXARGS	16		/* this seems like a lot.  Increase if needed... */
#define ARGLEN 80		/* again, this seems big enough... */

char func_name[TAGLEN];

good_c_func()    /* returns 1 if at a valid c func def, else 0 */
{
	int func_name_len;

	func_name_len = parse_string(1, "[a-zA-Z0-9_]+", func_name);
	if (!func_name_len || parse_string(-1, "^[ \t]*#[^#\n]*", NULL)
		|| !parse_string(1, "[ \t]*%(", NULL)
		|| !move_level(1, "(", ")"))
		return 0;       /* after identifier, skip over arg list */
	re_search(1, "([ \t\n]|(/<*>([^*]|<*>[^/])*<*>/)|//.*$)*");
	if (!parse_string(1, "[A-Za-z{]", NULL))
		return 0;       /* must not have one of these chars */
	return 1;
}

/*
 * Here is a command that will list all the c functions in the current buffer.
 */

command list_functions()
{
	int start = point;
	char *obufname = bufname;
	int end;

	iter = 0;
	point = 0;
	zap ("*Functions*");

	while (re_search(1, "({|/<*|/>|[a-zA-Z0-9_]+)"))
	{
		end = point;            /* find {, comment open, or ident */
		point = matchstart;
		if (curchar() == '{')
			skip_c_braces();
		else if (curchar() == '/')
		{
			if (character(point + 1) == '*')
			{
				point += 2;
				search(1, "*/");
			} else nl_forward();   /* c++ comment */
		}
		else if (good_c_func())
		{
			point = end;
			bufname = "*Functions*";
			bprintf ("%s()\n", func_name);
			bufname = obufname;
		}
	}

	point = start;
	view_buffer("*Functions*",0);
}


/*
 * Here is a command that will build a list of ANSI prototypes
 * for all the C functions in the current buffer. - AAC
 */

user int want_optional_arguments = 1;	/* defaults to wanting "function P((args));" format */
user int want_argument_names = 1;		/* defaults to wanting (type name, type name...) style args */
										/* instead of old non-ANSI style (type, type...) args */
char modifiers[14][10] =
{
	"signed",
	"unsigned",
	"long",
	"short",
	"register",
	"volatile",
	"const",
	"auto",
	"static",
	"extern",
	"struct",
	"union",
	""
};

int end_of_prior_funct = 0;

command make_prototypes()
{
	int start = point, preamble_count = 0, delim, indx, tindx, end, done, argptr = 0, fold_save;
	int len, func_start, string_start, string_end, area_start, area_end, type_end, first;
	char last_char, *obufname = bufname;
	char tmpstr[ARGLEN], preamble[ARGLEN];
	char arg_name[MAXARGS][ARGLEN], arg_type[MAXARGS][ARGLEN];	/* arrays to hold args parsed */

	fold_save = case_fold;
	case_fold = 0;							/* we need _EXACT_ r-e matches!! */
	iter = 0;
	point = 0;
	zap ("*Prototypes*");

	if (want_optional_arguments)			/* if we want P((s)) type params, define P(s) */
	{
		bufname = "*Prototypes*";
		stuff("#ifdef __STDC__\n#define P(s) s\n#else\n#define P(s) ()\n#endif\n\n");
		bufname = obufname;
	}

	while (re_search(1, "({|/<*|/>|[a-zA-Z0-9_]+)"))
	{
		end = point;						/* find {, comment open, or ident */
		point = matchstart;
		if (curchar() == '{')				/* if we find {, skip to ending } */
		{
			skip_c_braces();
			end_of_prior_funct = point;		/* make a note of this pos. */
		}
		else if (curchar() == '/')
		{									/* else skip comment(s) */
			if (character(point + 1) == '*')
			{
				point += 2;
				search(1, "*/");
			} else nl_forward();			/* skip c++ comments, too */
		}
		else if (good_c_func())				/* else check identifier for valid function */
		{
			point = end;					/* put point at end of func name */
			re_search(-1, "[ \t\n*]");		/* search backward to beginning of name */
			for (indx = 0; point > end_of_prior_funct; point--)	/* copy chars backward to tmpstr */
			{
				if (curchar() == ';' || curchar() == '}')	/* ...until we hit the first ; or } char... */
					break;
				else if (curchar() == '/')
				{											/* ...skipping comment(s)... */
					if (character(point - 1) == '*')
					{
						point -= 2;
						search(-1, "/*");
					} else nl_reverse();					/* ... skip c++ comments, too... */
					continue;
				}
				if (curchar() == '\n' || curchar() == '\t')	/* repl tabs or newlines w/spaces */
					 tmpstr[indx++] = ' ';
				else tmpstr[indx++] = curchar();
			}
			tmpstr[indx] = '\0';							/* terminate reversed string properly */

			last_char = ' ';								/* set reverser to ignore leading spaces */
			preamble_count = 0;
			for (indx = strlen(tmpstr) - 1; indx >= 0; indx--)	/* reverse tmpstr string to preamble */
			{
				if (last_char == ' ')						/* eliminate any multiple spaces */
				{
					if (tmpstr[indx] == last_char)
						continue;
				}
				last_char = preamble[preamble_count++] = tmpstr[indx];
			}
			preamble[preamble_count] = '\0';				/* terminate preamble (forward) string properly */

			if (preamble_count < 2)				/* if 0 or 1 - no chars copied to preamble */
				strcpy(preamble, "int ");		/* assume function is returning int by default */

			point = end;						/* point is after func name */
			while (curchar() != '(')			/* run up to ( */
				point++;
			point++;							/* 1 char past ( */
			func_start = delim = point;
			argptr = 0;							/* no args counted yet */
			while (argptr < MAXARGS)			/* scan up to ) char */
			{
				string_start = point;
				re_search(1, "[,)]");			/* search fwd for comma or close */
				delim = matchstart;				/* point at end of string of tokens */
				re_search(-1, "[A-Za-z0-9_]+");	/* search backward 1 token */
	
				if (matchend <= string_start)			/* if we are at or before the start of the string, */
				{										/* the token was a name w/no type identifier or none */
					parse_string(1, "[a-zA-Z0-9_]+", arg_name[argptr]);	/* copy token to arg name array */
					if (matchstart < func_start)		/* if match is outside the (, no valid tokens */
						break;
					if (!strcmp(arg_name[argptr], "void"))
						break;
					arg_type[argptr++][0] = '\0';		/* we have no type for this arg yet */
				}
				else
				{
					parse_string(1, "[a-zA-Z0-9_]+", arg_name[argptr]);	/* copy token to arg name array */
					point = string_start;				/* move back to start of whole string */
					for (indx = 0; point < matchstart; indx++)	/* copy everything up to token's name */
						arg_type[argptr][indx] = character(point++);
					arg_type[argptr++][indx] = '\0';	/* copy whole string of token(s) which is the type */
				}
				point = delim + 1;						/* move point to on top of delimiter */
				re_search(1, "[A-Za-z0-9_]");			/* find next non-whitespace char after delim */
				point--;
				if (curchar() == ')' || character(delim) == ')')	/* if we found a ), we are done */
					break;
			}
			if (argptr != MAXARGS)						/* if we didn't fill the list */
				arg_name[argptr][0] = '\0';				/* terminate arg list w/NULL name */

			point = func_start;							/* put point back a start of func (1 past '(') */
			re_search(1, "[)]");						/* put point 1 past ending ) after function decl. */
			area_start = point;							/* save this position */
			re_search(1, "[{]");						/* find beginning { after function decl. */
			area_end = matchstart;						/* end of search area is at the opening brace */
			
			for (indx = 0; indx < MAXARGS && arg_name[indx][0]; indx++)	/* for each arg in array until a NULL arg */
			{
				if (arg_type[indx][0])					/* loop if there is an arg type there */
					continue;
				if (area_end - area_start < 4)			/* if no formal params, any args w/o types must be ints */
				{
					strcpy(arg_type[indx], "int ");		/* we always default to ints in C-land */
					continue;
				}
				point = area_start;						/* start looking for name at start of area */
				re_search(1, arg_name[indx]);			/* find arg name below function decl. */
				point = string_end = matchstart;		/* note the place we found it at */
				re_search(-1, "[;)]");					/* search backward for prior ) or ; char */
				string_start = point;					/* note position of start of arg data type param */
				point++;								/* move past the found delimiter */
				done = 0;
				first = 1;
				arg_type[indx][0] = '\0';				/* terminate type for upcoming concatenation */
				while (!done)			/* collect tokens until we get an unknown one (the actual type) */
				{
					for (len = 0; !len && point <= area_end; point++)
						len = parse_string(1, "[a-zA-Z0-9_]+", tmpstr);	/* find next token, arg data type (?) */
					type_end = matchend;
					strcat(arg_type[indx], tmpstr);		/* concatenate token to arg_type[] */
					strcat(arg_type[indx], " ");		/* tack on a space separating type from any extras */
 					if (first && (matchstart >= string_end)) /* if we find it's on or past the name, 1st time only, */
					{
						type_end = string_start;
						strcpy(arg_type[indx], "int ");	/* default arg to int if type is not found */
						break;
					}
					first = 0;							/* now on 2nd-nth token */
					done = 1;							/* unless we find a match next, we're done */
					for (tindx = 0; modifiers[tindx][0]; tindx++)	/* scan modifiers table */
					{
						if (!strcmp(tmpstr, modifiers[tindx]))	/* if we found a valid modifier */
						{
							done = 0;					/* go collect another token */
							break;						/* else it was int, char, long, etc. or a struct name */
						}
					}
				}
				point = string_end;						/* move back to string end again */
				re_search(-1, "[);,]");					/* search backward for delimiter */
				if (point == string_start)				/* if we're back at the start, we didn't find a comma, */
					point = type_end;					/* so position point after end of type word */
				re_search(1, "[A-Za-z0-9_*]");			/* find next non-whitespace char after point delim */
				point--;
				tmpstr[1] = '\0';						/* terminate a 1 character long string */
				while (point < string_end)
				{
					tmpstr[0] = character(point++);
					strcat(arg_type[indx], tmpstr);		/* tack on chars until we get up to the name */
				}
			}

			point = end;								/* restore point to after func name */
			bufname = "*Prototypes*";					/* switch to proto buffer */
			if (want_optional_arguments)
				 bprintf ("%s%s P((", preamble, func_name);
			else bprintf ("%s%s(", preamble, func_name);
			if (!arg_name[0][0])						/* if no args in array at all, */
				bprintf("void");						/* make function args (void) */
			else
			{
				for (indx = 0; indx < MAXARGS && arg_name[indx][0]; indx++) /* for each arg in array until NULL arg */
				{
					if (!arg_type[indx][0])				/* quit if there is no arg type there */
						break;							/* (this should never happen...) */
					bprintf("%s", arg_type[indx]);
					if (want_argument_names)			/* old style prototypes list just types, not names */
						bprintf("%s", arg_name[indx]);
					if (arg_name[indx + 1][0])			/* if there are more args */
						bprintf(", ");					/* add comma and space */
				}
			}
			if (want_optional_arguments)
				 bprintf ( "));\n");
			else bprintf ( ");\n");

			bufname = obufname;
		}
	}

	if (want_optional_arguments)						/* now undefine P() after we're done using it */
	{
		bufname = "*Prototypes*";
		stuff("\n#undef P\n");
		bufname = obufname;
	}

	bufname = "*Prototypes*";
	mark = 0;
	point = size();
	kill_region();
	yank();
	bufname = obufname;

	case_fold = fold_save;
	point = start;
	view_buffer("*Prototypes*",0);
}
